from operator import itemgetter
import re
from typing import (
    Any,
    Dict,
    Iterable,
    NamedTuple,
    Optional,
    List,
    Tuple,
    TYPE_CHECKING,
    Union,
)
from typing_extensions import Literal

if TYPE_CHECKING:  # pragma: no cover
    from .console import (
        Console,
        ConsoleOptions,
        JustifyValues,
        RenderResult,
        RenderableType,
    )

from .containers import Lines
from .style import Style
from .segment import Segment
from .render_width import RenderWidth
from ._tools import iter_last, iter_first_last
from ._wrap import divide_line


class Span(NamedTuple):
    """A marked up region in some text."""

    start: int
    end: int
    style: Union[str, Style]

    def __repr__(self) -> str:
        return f'<span {self.start}:{self.end} "{self.style}">'

    def __bool__(self) -> bool:
        return self.end > self.start

    def split(self, offset: int) -> Tuple["Span", Optional["Span"]]:
        """Split a span in to 2 from a given offset."""

        if offset < self.start:
            return self, None
        if offset >= self.end:
            return self, None

        start, end, style = self
        span1 = Span(start, min(end, offset), style)
        span2 = Span(span1.end, end, style)
        return span1, span2

    def move(self, offset: int) -> "Span":
        """Move start and end by a given offset.
        
        Args:
            offset (int): Number of characters to add to start and end.
        
        Returns:
            TextSpan: A new TextSpan with adjusted position.
        """
        start, end, style = self
        return Span(start + offset, end + offset, style)

    def right_crop(self, offset: int) -> "Span":
        """Crop the span at the given offset.
        
        Args:
            offset (int): A value between start and end.
        
        Returns:
            Span: A new (possibly smaller) span.
        """
        start, end, style = self
        if offset >= end:
            return self
        return Span(start, min(offset, end), style)


class Text:
    r"""Text with color / style.
        
        Args:
            text (str, optional): Default unstyled text. Defaults to "".
            style (Union[str, Style], optional): Base style for text. Defaults to "".
            justify (str, optional): Default alignment for text, "left", "center", "full" or "right". Defaults to None.
            end (str, optional): Character to end text with. Defaults to "\n".
    """

    def __init__(
        self,
        text: str = "",
        style: Union[str, Style] = "",
        justify: "JustifyValues" = None,
        end: str = "\n",
    ) -> None:

        self._text: List[str] = [text] if text else []
        self.style = style
        self.justify = justify
        self.end = end
        self._spans: List[Span] = []
        self._length: int = len(text)

    def __len__(self) -> int:
        return self._length

    def __bool__(self) -> bool:
        return bool(self._length)

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f"<text {self.text!r} {self._spans!r}>"

    def __add__(self, other: Any) -> "Text":
        if isinstance(other, (str, Text)):
            result = self.copy()
            result.append(other)
            return result
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Text):
            return NotImplemented
        return self.text == other.text and self._spans == other._spans

    @classmethod
    def from_markup(cls, text: str, style: Union[str, Style] = "") -> "Text":
        """Create Text instance from markup.
        
        Args:
            text (str): A string containing console markup.
        
        Returns:
            Text: A Text instance with markup rendered.
        """
        from .markup import render

        return render(text, style)

    @classmethod
    def assemble(
        cls,
        *parts: Tuple[str, Union[str, Style]],
        style: Union[str, Style] = "",
        justify: "JustifyValues" = None,
        end: str = "\n",
    ) -> "Text":
        text = cls(style=style, justify=justify, end=end)
        append = text.append
        for part in parts:
            append(*part)
        return text

    @property
    def text(self) -> str:
        """Get the text as a single string."""
        if len(self._text) not in (0, 1):
            text = "".join(self._text)
            self._text[:] = [text]
        return self._text[0] if self._text else ""

    @text.setter
    def text(self, new_text: str) -> "Text":
        """Set the text to a new value."""
        self._text[:] = [new_text]
        old_length = self._length
        self._length = len(new_text)
        if old_length > self._length:
            self._trim_spans()
        return self

    def copy(self) -> "Text":
        """Return a copy of this instance."""
        self.text
        copy_self = Text(
            self.text, style=self.style, justify=self.justify, end=self.end,
        )
        copy_self._spans[:] = self._spans[:]
        return copy_self

    def stylize(self, start: int, end: int, style: Union[str, Style]) -> None:
        """Apply a style to a portion of the text.
        
        Args:
            start (int): Start offset.
            end (int): End offset.
            style (str): Style name to apply.
        
        Returns:
            None: 
        """
        length = len(self)
        if end < 0 or start > length:
            # span not in range
            return
        self._spans.append(Span(max(0, start), min(length, end), style))

    def highlight_regex(
        self, re_highlight: str, style: Union[str, Style] = None
    ) -> int:
        """Highlight text with a regular expression, where group names are 
        translated to styles.
        
        Args:
            re_highlight (str): A regular expression            
        
        Returns:
            int: Number of regex matches
        """
        count = 0
        append_span = self._spans.append
        _Span = Span
        for match in re.finditer(re_highlight, self.text):
            _span = match.span
            if style:
                start, end = _span()
                append_span(_Span(start, end, style))
            count += 1
            for name, _ in match.groupdict().items():
                start, end = _span(name)
                if start != -1:
                    append_span(_Span(start, end, name))
        return count

    def highlight_words(
        self,
        words: Iterable[str],
        style: Union[str, Style],
        case_sensitive: bool = True,
    ) -> int:
        """Highlight words with a style.
        
        Args:
            words (Iterable[str]): Worlds to highlight.
            style (Union[str, Style]): Style to apply.
        
        Returns:
            int: Number of words highlighted.
        """
        re_words = "|".join(re.escape(word) for word in words)
        add_span = self._spans.append
        count = 0
        _Span = Span
        for match in re.finditer(
            re_words, self.text, flags=0 if case_sensitive else re.IGNORECASE
        ):
            start, end = match.span(0)
            add_span(_Span(start, end, style))
            count += 1
        return count

    def rstrip(self) -> None:
        """Trip whitespace from end of text."""
        self.text = self.text.rstrip()

    def set_length(self, new_length: int) -> None:
        """Set new length of the text, clipping or padding is required."""
        length = len(self)
        if length == new_length:
            return
        if length < new_length:
            self.pad_right(new_length - length)
        else:
            text = self.text[:new_length]
            self.text = text
            new_spans = []
            for span in self._spans:
                if span.start < new_length:
                    new_spans.append(span.right_crop(new_length))
            self._spans[:] = new_spans

    def __console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> Iterable[Segment]:
        lines = self.wrap(options.max_width, justify=self.justify or options.justify)
        all_lines = Text("\n").join(lines)
        yield from self._render_line(all_lines, console, options)

    def __console_width__(self, max_width: int) -> RenderWidth:
        text = self.text
        if not text.strip():
            return RenderWidth(len(text), len(text))
        max_text_width = max(len(line) for line in text.splitlines())
        min_text_width = max(len(word) for word in text.split())
        return RenderWidth(min_text_width, max_text_width)

    def _render_line(
        self, line: "Text", console: "Console", options: "ConsoleOptions"
    ) -> Iterable["Segment"]:
        """Render the rich text to the console.
        
        Args:
            console (Console): Console instance.
            options (ConsoleOptions): Console options.
        
        Returns:
            Iterable[Segment]: Result of render that may be written to the console.
        """

        text = line.text
        style_map: Dict[int, Style] = {}
        null_style = Style()

        def get_style(style: Union[str, Style]) -> Style:
            return console.get_style(style, default=null_style)

        enumerated_spans = list(enumerate(line._spans, 1))

        style_map = {index: get_style(span.style) for index, span in enumerated_spans}
        style_map[0] = get_style(self.style)

        spans = [
            (0, False, 0),
            *((span.start, False, index) for index, span in enumerated_spans),
            *((span.end, True, index) for index, span in enumerated_spans),
            (len(text), True, 0),
        ]
        spans.sort(key=itemgetter(0, 1))

        stack: List[int] = []
        stack_append = stack.append
        stack_pop = stack.remove

        _Segment = Segment

        style_cache: Dict[Tuple[int, ...], Style] = {}
        combine = Style.combine

        def get_current_style() -> Style:
            """Construct current style from stack."""
            style_ids = tuple(sorted(stack))
            cached_style = style_cache.get(style_ids)
            if cached_style is not None:
                return cached_style
            current_style = combine(style_map[_style_id] for _style_id in style_ids)
            style_cache[style_ids] = current_style
            return current_style

        for (offset, leaving, style_id), (next_offset, _, _) in zip(spans, spans[1:]):
            if leaving:
                stack_pop(style_id)
            else:
                stack_append(style_id)
            if next_offset > offset:
                yield _Segment(text[offset:next_offset], get_current_style())
        if self.end:
            yield _Segment(self.end)

    def join(self, lines: Iterable["Text"]) -> "Text":
        """Join text together with this instance as the separator.
        
        Args:
            lines (Iterable[Text]): An iterable of Text instances to join.
        
        Returns:
            Text: A new text instance containing join text.
        """

        new_text = Text()
        append = new_text.append
        for last, line in iter_last(lines):
            append(line)
            if not last:
                append(self)
        return new_text

    def _trim_spans(self) -> None:
        """Remove or modify any spans that are over the end of the text."""
        new_length = self._length
        spans: List[Span] = []
        append = spans.append
        for span in self._spans:
            if span.end < new_length:
                append(span)
                continue
            if span.start >= new_length:
                continue
            append(span.right_crop(new_length))
        self._spans[:] = spans

    def pad_left(self, count: int, character: str = " ") -> None:
        """Pad the left with a given character.
        
        Args:
            count (int): Number of characters to pad.
            character (str, optional): Character to pad with. Defaults to " ".
        """
        assert len(character) == 1, "Character must be a string of length 1"
        if count:
            self.text = f"{character * count}{self.text}"
            self._spans[:] = [span.move(count) for span in self._spans]

    def pad_right(self, count: int, character: str = " ") -> None:
        """Pad the right with a given character.
        
        Args:
            count (int): Number of characters to pad.
            character (str, optional): Character to pad with. Defaults to " ".
        """
        assert len(character) == 1, "Character must be a string of length 1"
        if count:
            self.text = f"{self.text}{character * count}"

    def append(
        self, text: Union["Text", str], style: Union[str, "Style"] = None
    ) -> None:
        """Add text with an optional style.
        
        Args:
            text (Union[Text, str]): A str or Text to append.
            style (str, optional): A style name. Defaults to None.
        """
        if isinstance(text, str):
            self._text.append(text)
            offset = len(self)
            text_length = len(text)
            if style is not None:
                self._spans.append(Span(offset, offset + text_length, style))
            self._length += text_length
        elif isinstance(text, Text):
            if style is not None:
                raise ValueError("style must not be set when appending Text instance")
            text_length = self._length
            if text.style is not None:
                self._spans.append(
                    Span(text_length, text_length + len(text), text.style),
                )
            self._text.append(text.text)
            self._spans.extend(
                Span(start + text_length, end + text_length, style)
                for start, end, style in text._spans
            )
            self._length += len(text)
        else:
            raise TypeError("Only str or Text can be appended to Text")

    def split(self, separator="\n", include_separator: bool = False) -> Lines:
        r"""Split rich text in to lines, preserving styles.
        
        Args:
            separator (str, optional): String to split on. Defaults to "\n".
        
        Returns:
            List[RichText]: A list of rich text, one per line of the original.
        """
        assert separator, "separator must not be empty"

        text = self.text
        if separator not in text:
            return Lines([self.copy()])
        if text.endswith(separator):
            text = text[: -len(separator)]
        offsets: List[int] = []
        append = offsets.append
        offset = 0
        while True:
            try:
                offset = text.index(separator, offset) + len(separator)
            except ValueError:
                break
            append(offset)
        lines = self.divide(offsets)
        if not include_separator:
            separator_length = len(separator)
            for line in lines:
                if line.text.endswith(separator):
                    line.right_crop(separator_length)
        return lines

    def divide(self, offsets: Iterable[int]) -> Lines:
        """Divide text in to a number of lines at given offsets.

        Args:
            offsets (Iterable[int]): Offsets used to divide text.
        
        Returns:
            Lines: New RichText instances between offsets.
        """

        if not offsets:
            line = self.copy()
            return Lines([line])

        text = self.text
        text_length = len(text)
        divide_offsets = [0, *offsets, text_length]
        line_ranges = list(zip(divide_offsets, divide_offsets[1:]))
        average_line_length = -(-text_length // len(line_ranges))

        new_lines = Lines(
            Text(text[start:end], style=self.style, justify=self.justify)
            for start, end in line_ranges
        )
        line_ranges = [
            (offset, offset + len(line))
            for offset, line in zip(divide_offsets, new_lines)
        ]

        for span in self._spans:
            line_index = span.start // average_line_length

            line_start, line_end = line_ranges[line_index]
            if span.start < line_start:
                while True:
                    line_index -= 1
                    line_start, line_end = line_ranges[line_index]
                    if span.end >= line_start:
                        break
            elif span.start > line_end:
                while True:
                    line_index += 1
                    line_start, line_end = line_ranges[line_index]
                    if span.start <= line_end:
                        break

            while True:
                span, new_span = span.split(line_end)
                new_lines[line_index]._spans.append(span.move(-line_start))
                if new_span is None:
                    break
                span = new_span
                line_index += 1
                line_start, line_end = line_ranges[line_index]

        return new_lines

    def right_crop(self, amount: int = 1) -> None:
        """Remove a number of characters from the end of the text."""
        self.text = self.text[:-amount]

    def wrap(self, width: int, justify: "JustifyValues" = "left") -> Lines:
        """Word wrap the text.
        
        Args:
            width (int): Number of characters per line.
            justify (bool, optional): True to pad lines with spaces. Defaults to False.
        
        Returns:
            Lines: Number of lines.
        """

        lines: Lines = Lines()
        for line in self.split():
            offsets = divide_line(str(line), width)
            new_lines = line.divide(offsets)
            if justify:
                new_lines.justify(width, align=justify)
            lines.extend(new_lines)
        return lines

    def fit(self, width: int) -> Lines:
        """Fit the text in to given width by chopping in to lines.
        
        Args:
            width (int): Maximum characters in a line.
        
        Returns:
            Lines: List of lines.
        """
        lines: Lines = Lines()
        append = lines.append
        for line in self.split():
            line.set_length(width)
            append(line)
        return lines


if __name__ == "__main__":  # pragma: no cover
    from .console import Console

    console = Console()
    text = Text(
        """Hello, self! hello World

hello  worhello\n"""
    )
    text.highlight_regex("self", "reverse")
    console.print(text)




from operator import itemgetter
from typing import Iterable, NamedTuple, TYPE_CHECKING

from . import errors
from .segment import Segment

if TYPE_CHECKING:
    from .console import RenderableType


class RenderWidth(NamedTuple):
    """Range of widths for a renderable object."""

    minimum: int
    maximum: int

    @property
    def span(self) -> int:
        """Get difference between maximum and minimum."""
        return self.maximum - self.minimum

    def normalize(self) -> "RenderWidth":
        minimum, maximum = self
        minimum = max(0, minimum)
        return RenderWidth(minimum, max(minimum, maximum))

    def with_maximum(self, width: int) -> "RenderWidth":
        """Get a RenderableWith where the widths are <= width.
        
        Args:
            width (int): Maximum desired width.
        
        Returns:
            RenderableWidth: new RenderableWidth object.
        """
        minimum, maximum = self
        return RenderWidth(min(minimum, width), min(maximum, width))

    @classmethod
    def get(cls, renderable: "RenderableType", max_width: int) -> "RenderWidth":
        """Get desired width for a renderable."""
        if hasattr(renderable, "__console__"):
            get_console_width = getattr(renderable, "__console_width__", None)
            if get_console_width is not None:
                render_width = get_console_width(max_width).with_maximum(max_width)
                return render_width.normalize()
            else:
                return RenderWidth(1, max_width)
        elif isinstance(renderable, Segment):
            text, _style = renderable
            width = min(max_width, len(text))
            return RenderWidth(width, width)
        elif isinstance(renderable, str):
            text = renderable.rstrip()
            return RenderWidth(len(text), len(text))
        else:
            raise errors.NotRenderableError(
                f"Unable to get render width for {renderable!r}; "
                "a str, Segment, or object with __console__ method is required"
            )

    @classmethod
    def measure(
        cls, renderables: Iterable["RenderableType"], max_width: int
    ) -> "RenderWidth":
        """Measure a number of renderables."""

        render_widths = [
            RenderWidth.get(renderable, max_width) for renderable in renderables
        ]
        measured_width = RenderWidth(
            max(render_widths, key=itemgetter(0)).minimum,
            max(render_widths, key=itemgetter(1)).maximum,
        )
        return measured_width



import re
from typing import Match, Union

from .console import Console, ConsoleOptions, ConsoleRenderable, RenderResult
from .segment import Segment
from .style import Style
from ._emoji_codes import EMOJI
from ._emoji_replace import _emoji_replace


class NoEmoji(Exception):
    """No emoji by that name."""


class Emoji:
    __slots__ = ["name", "style", "_char"]

    def __init__(self, name: str, style: Union[str, Style] = "none") -> None:
        """A single emoji character.
        
        Args:
            name (str): Name of emoji.
            style (Union[str, Style], optional): Optional style. Defaults to None.
        
        Raises:
            NoEmoji: If the emoji doesn't exist.
        """
        self.name = name
        self.style = style
        try:
            self._char = EMOJI[name]
        except KeyError:
            raise NoEmoji(f"No emoji called {name!r}")

    @classmethod
    def replace(cls, text: str) -> str:
        """Replace emoji markup with coresponding unicode characters.
        
        Args:
            text (str): A string with emojis codes, e.g. "Hello :smiley:!"
        
        Returns:
            str: A string with emoji codes replaces with actual emoji.
        """
        return _emoji_replace(text)

    def __repr__(self) -> str:
        return f"<emoji {self.name!r}>"

    def __str__(self) -> str:
        return self._char

    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        yield Segment(self._char, console.get_style(self.style))


if __name__ == "__main__":  # pragma: no cover
    from .console import Console

    c = Console(markup=False)

    e = Emoji("thumbs_up")
    print(repr(e))
    print(e)

    c.print(Emoji("thumbs_up"))
    # c.print("Hello")
    c.print("Hello World")
    from .panel import Panel

    c.print(Panel(Emoji.replace("Hello, :smiley: ! :thumbs_up: :sfwdfwer:")))

    c.print("Here is a :smiley:  :link: Hello World")

    c.print(":beer:")



import re
from typing import Iterable, List, Tuple

re_word = re.compile(r"\s*\S+\s*")


def words(text: str) -> Iterable[Tuple[int, str]]:
    position = 0
    word_match = re_word.match(text, position)
    while word_match is not None:
        start, position = word_match.span()
        word = word_match.group(0)
        yield start, word
        word_match = re_word.match(text, position)


def divide_line(text: str, width: int) -> List[int]:
    divides: List[int] = []
    append = divides.append
    line_size = 0
    for position, word in words(text):
        if line_size + len(word.rstrip()) > width:
            if position:
                append(position)
            line_size = 0
        line_size += len(word)
    return divides



from colorama import init

from typing import Any, IO, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .console import Console

# Global console used by alternative print
_console: Optional["Console"] = None


def print(
    *objects: Any,
    sep=" ",
    end="\n",
    file: IO[str] = None,
    flush: bool = False,
    log_locals: bool = False
):
    global _console
    if _console is None:
        from .console import Console

        _console = Console(log_time=False)

    write_console = _console if file is None else Console(log_time=False, file=file)
    return write_console.log(
        *objects, sep=sep, end=end, log_locals=log_locals, _stack_offset=2
    )


init()

if __name__ == "__main__":  # pragma: no cover
    print("Hello, **World**", log_locals=True)



import configparser
from typing import Dict, IO

from .style import Style


class Theme:
    def __init__(self, styles: Dict[str, Style] = None):
        self.styles = styles or {}

    @property
    def config(self) -> str:
        """Get contents of a config file for this theme."""
        config_lines = ["[styles]"]
        append = config_lines.append
        for name, style in sorted(self.styles.items()):
            append(f"{name} = {style}")
        config = "\n".join(config_lines)
        return config

    @classmethod
    def from_file(cls, config_file: IO[str], source: str = None) -> "Theme":
        config = configparser.ConfigParser()
        config.read_file(config_file, source=source)
        styles = {name: Style.parse(value) for name, value in config.items("styles")}
        theme = Theme(styles)
        return theme

    @classmethod
    def read(cls, path: str) -> "Theme":
        with open(path, "rt") as config_file:
            return cls.from_file(config_file, source=path)



from math import sqrt
from functools import lru_cache
from typing import List, Sequence, Tuple

from .color_triplet import ColorTriplet


class Palette:
    """A palette of available colors."""

    def __init__(self, colors: Sequence[Tuple[int, int, int]]):
        self._colors = colors

    def __getitem__(self, number: int) -> ColorTriplet:
        return ColorTriplet(*self._colors[number])

    @property
    def hex(self) -> List[str]:
        """Get a list of hex encoded colors."""
        return [f"{red:02x}{green:02x}{blue:02x}" for red, green, blue in self._colors]

    # This is somewhat inefficient and needs caching
    @lru_cache(maxsize=1000)
    def match(self, color: Tuple[int, int, int]) -> int:
        """Find a color from a palette that most closely matches a given color"""
        red1, green1, blue1 = color
        _sqrt = sqrt

        def get_color_distance(color: Tuple[int, int, int]) -> float:
            """Get the distance to a color."""
            red2, green2, blue2 = color
            red_mean = int((red1 + red2) / 2)
            red = red1 - red2
            green = green1 - green2
            blue = blue1 - blue2
            return _sqrt(
                (((512 + red_mean) * red * red) >> 8)
                + 4 * green * green
                + (((767 - red_mean) * blue * blue) >> 8)
            )

        min_index, _min_color = min(
            enumerate(self._colors), key=lambda _color: get_color_distance(_color[1]),
        )
        return min_index




import logging
from datetime import datetime
from logging import Handler, LogRecord
from pathlib import Path

from rich._log_render import LogRender
from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.markup import render
from rich.text import Text

# LOG_LEVELS = ["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


class RichHandler(Handler):
    """A logging handler with rich output."""

    KEYWORDS = ["GET", "POST", "HEAD", "PUT", "DELETE", "OPTIONS", "TRACE", "PATCH"]

    def __init__(self, level: int = logging.NOTSET) -> None:
        super().__init__(level=level)
        self.console = Console()
        self.highlighter = ReprHighlighter()
        self._log_render = LogRender(show_level=True)

    def emit(self, record: LogRecord) -> None:
        path = Path(record.pathname).name
        log_style = f"logging.level.{record.levelname.lower()}"
        message = self.format(record)
        time_format = None if self.formatter is None else self.formatter.datefmt
        log_time = datetime.fromtimestamp(record.created)

        level = Text()
        level.append(record.levelname, log_style)
        message_text = Text(message)
        message_text.highlight_words(self.KEYWORDS, "logging.keyword")
        message_text = self.highlighter(message_text)

        self.console.print(
            self._log_render(
                self.console,
                [message_text],
                log_time=log_time,
                time_format=time_format,
                level=level,
                path=path,
                line_no=record.lineno,
            )
        )


if __name__ == "__main__":  # pragma: no cover
    from time import sleep

    FORMAT = "%(message)s"
    # FORMAT = "%(asctime)-15s - %(level) - %(message)s"
    logging.basicConfig(
        level="NOTSET", format=FORMAT, datefmt="[%X] ", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")

    log.info("Server starting...")
    log.info("Listening on http://127.0.0.1:8080")
    sleep(1)

    log.info("GET /index.html 200 1298")
    log.info("GET /imgs/backgrounds/back1.jpg 200 54386")
    log.info("GET /css/styles.css 200 54386")
    log.warning("GET /favicon.ico 404 242")
    sleep(1)

    log.debug(
        "JSONRPC request\n--> %r\n<-- %r",
        {
            "version": "1.1",
            "method": "confirmFruitPurchase",
            "params": [["apple", "orange", "mangoes", "pomelo"], 1.123],
            "id": "194521489",
        },
        {"version": "1.1", "result": True, "error": None, "id": "194521489"},
    )
    log.error("Unable to find 'pomelo' in database!")
    log.info("POST /jsonrpc/ 200 65532")
    log.info("POST /admin/ 401 42234")
    log.warning("password was rejected for admin site.")
    try:
        1 / 0
    except:
        log.exception("An error of some kind occurred!")
    sleep(1)
    log.critical("Out of memory!")
    log.info("Server exited with code=-1")



from dataclasses import dataclass
from typing import Any, ClassVar, Dict, Iterable, List, Optional, Type, Union

from commonmark.blocks import Parser

from . import box
from .console import (
    Console,
    ConsoleOptions,
    ConsoleRenderable,
    JustifyValues,
    RenderResult,
    Segment,
)
from .containers import Renderables
from .panel import Panel
from .rule import Rule
from .style import Style, StyleStack
from .syntax import Syntax
from .text import Lines, Text
from ._stack import Stack
from ._tools import iter_first, iter_first_last


class MarkdownElement:

    new_line: ClassVar[bool] = True

    @classmethod
    def create(cls, markdown: "Markdown", node: Any) -> "MarkdownElement":
        """Factory to create markdown element,
        
        Args:
            markdown (Markdown): THe parent Markdown object.
            node (Any): A node from Pygments.
        
        Returns:
            MarkdownElement: A new markdown element
        """
        return cls()

    def on_enter(self, context: "MarkdownContext"):
        """Called when the node is entered.
        
        Args:
            context (MarkdownContext): The markdown context.
        """

    def on_text(self, context: "MarkdownContext", text: str) -> None:
        """Called when text is parsed.
        
        Args:
            context (MarkdownContext): The markdown context.
        """

    def on_leave(self, context: "MarkdownContext") -> None:
        """Called when the parser leaves the element.
        
        Args:
            context (MarkdownContext): [description]
        """

    def on_child_close(
        self, context: "MarkdownContext", child: "MarkdownElement"
    ) -> bool:
        """Called when a child element is closed.

        This method allows a parent element to take over rendering of its children.
        
        Args:
            context (MarkdownContext): The markdown context.
            child (MarkdownElement): The child markdown element.
        
        Returns:
            bool: Return True to render the element, or False to not render the element.
        """
        return True

    def __console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> "RenderResult":
        return ()


class UnknownElement(MarkdownElement):
    """An unknown element.
    
    Hopefully there will be no unknown elements, and we will have a MarkdownElement for
    everything in the document.
    
    """


class TextElement(MarkdownElement):
    """Base class for elements that render text."""

    style_name = "none"

    def on_enter(self, context: "MarkdownContext") -> None:
        self.style = context.enter_style(self.style_name)
        self.text = Text(justify="left")

    def on_text(self, context: "MarkdownContext", text: str) -> None:
        self.text.append(text, context.current_style)

    def on_leave(self, context: "MarkdownContext") -> None:
        context.leave_style()

    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        yield self.text


class Paragraph(TextElement):
    """A Paragraph."""

    style_name = "markdown.paragraph"
    justify: JustifyValues

    @classmethod
    def create(cls, markdown: "Markdown", node) -> "Paragraph":
        return cls(justify=markdown.justify or "left")

    def __init__(self, justify: JustifyValues) -> None:
        self.justify = justify

    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        self.text.justify = self.justify
        yield self.text


class Heading(TextElement):
    """A heading."""

    @classmethod
    def create(cls, markdown: "Markdown", node: Any) -> "Heading":
        heading = Heading(node.level)
        return heading

    def on_enter(self, context: "MarkdownContext") -> None:
        self.text = Text()
        context.enter_style(self.style_name)

    def __init__(self, level: int) -> None:
        self.level = level
        self.style_name = f"markdown.h{level}"
        super().__init__()

    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        text = self.text
        text.justify = "center"
        if self.level == 1:
            # Draw a border around h1s
            yield Panel(
                text, box=box.DOUBLE, style="markdown.h1.border",
            )
        else:
            # Styled text for h2 and beyond
            yield text


class CodeBlock(TextElement):
    """A code block with syntax highlighting."""

    style_name = "markdown.code_block"

    @classmethod
    def create(cls, markdown: "Markdown", node: Any) -> "CodeBlock":
        if node.info is None:
            return cls("default", markdown.code_theme)
        lexer_name, _, _ = node.info.partition(" ")
        return cls(lexer_name, markdown.code_theme)

    def __init__(self, lexer_name: str, theme: str) -> None:
        self.lexer_name = lexer_name
        self.theme = theme

    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        code = str(self.text).rstrip()
        syntax = Syntax(code, self.lexer_name, theme=self.theme)
        yield syntax


class BlockQuote(TextElement):
    """A block quote."""

    style_name = "markdown.block_quote"

    def __init__(self) -> None:
        self.elements: Renderables = Renderables()

    def on_child_close(
        self, context: "MarkdownContext", child: "MarkdownElement"
    ) -> bool:
        self.elements.append(child)
        return False

    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        # Render surrounded by quotes.
        render_options = options.update(width=options.max_width - 4)
        lines = console.render_lines(self.elements, render_options, style=self.style)

        style = self.style
        new_line = Segment("\n")
        left_quote = Segment("“ ", style)
        right_quote = Segment(" ”", style)
        padding = Segment("  ", style)

        for first, last, line in iter_first_last(lines):
            yield left_quote if first else padding
            yield from line
            yield right_quote if last else padding
            yield new_line


class HorizontalRule(MarkdownElement):
    """A horizontal rule to divide secions."""

    new_line = False

    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        style = console.get_style("markdown.hr")
        yield Rule(style=style)


class ListElement(MarkdownElement):
    """A list element."""

    @classmethod
    def create(cls, markdown: "Markdown", node: Any) -> "ListElement":
        list_data = node.list_data
        return cls(list_data["type"], list_data["start"])

    def __init__(self, list_type: str, list_start: Optional[int]) -> None:
        self.items: List[ListItem] = []
        self.list_type = list_type
        self.list_start = list_start

    def on_child_close(
        self, context: "MarkdownContext", child: "MarkdownElement"
    ) -> bool:
        assert isinstance(child, ListItem)
        self.items.append(child)
        return False

    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        if self.list_type == "bullet":
            for item in self.items:
                yield from item.render_bullet(console, options)
        else:
            number = 1 if self.list_start is None else self.list_start
            last_number = number + len(self.items)
            for item in self.items:
                yield from item.render_number(console, options, number, last_number)
                number += 1


class ListItem(TextElement):
    """An item in a list."""

    style_name = "markdown.item"

    def __init__(self) -> None:
        self.elements: Renderables = Renderables()

    def on_child_close(
        self, context: "MarkdownContext", child: "MarkdownElement"
    ) -> bool:
        self.elements.append(child)
        return False

    def render_bullet(self, console: Console, options: ConsoleOptions) -> RenderResult:
        render_options = options.update(width=options.max_width - 3)
        lines = console.render_lines(self.elements, render_options, style=self.style)
        bullet_style = console.get_style("markdown.item.bullet")

        bullet = Segment(" • ", bullet_style)
        padding = Segment(" " * 3, bullet_style)
        new_line = Segment("\n")
        for first, line in iter_first(lines):
            yield bullet if first else padding
            yield from line
            yield new_line

    def render_number(
        self, console: Console, options: ConsoleOptions, number: int, last_number: int
    ) -> RenderResult:
        number_width = len(str(last_number)) + 2
        render_options = options.update(width=options.max_width - number_width)
        lines = console.render_lines(self.elements, render_options, style=self.style)
        number_style = console.get_style("markdown.item.number")

        new_line = Segment("\n")
        padding = Segment(" " * number_width, number_style)
        numeral = Segment(f"{number}".rjust(number_width - 1) + " ", number_style)
        for first, line in iter_first(lines):
            yield numeral if first else padding
            yield from line
            yield new_line


class MarkdownContext:
    """Manages the console render state."""

    def __init__(self, console: Console, options: ConsoleOptions, style: Style) -> None:
        self.console = console
        self.options = options
        self.style_stack: StyleStack = StyleStack(style)
        self.stack: Stack[MarkdownElement] = Stack()

    @property
    def current_style(self) -> Style:
        """Current style which is the product of all styles on the stack."""
        return self.style_stack.current

    @property
    def width(self) -> int:
        """The width of the console."""
        return self.options.max_width

    def on_text(self, text: str) -> None:
        """Called when the parser visits text."""
        self.stack.top.on_text(self, text)

    def enter_style(self, style_name: Union[str, Style]) -> Style:
        """Enter a style context."""
        style = self.console.get_style(style_name)
        self.style_stack.push(style)
        return self.current_style

    def leave_style(self) -> Style:
        """Leave a style context."""
        style = self.style_stack.pop()
        return style


class Markdown:

    elements: ClassVar[Dict[str, Type[MarkdownElement]]] = {
        "paragraph": Paragraph,
        "heading": Heading,
        "code_block": CodeBlock,
        "block_quote": BlockQuote,
        "thematic_break": HorizontalRule,
        "list": ListElement,
        "item": ListItem,
    }
    inlines = {"emph", "strong", "code", "link"}

    def __init__(
        self,
        markup: str,
        code_theme: str = "monokai",
        justify: JustifyValues = None,
        style: Union[str, Style] = "none",
    ) -> None:
        """Parses the markup."""
        self.markup = markup
        parser = Parser()
        self.parsed = parser.parse(markup)
        self.code_theme = code_theme
        self.justify = justify
        self.style = style

    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        """Render markdown to the console."""
        style = console.get_style(self.style)
        context = MarkdownContext(console, options, style)
        nodes = self.parsed.walker()
        inlines = self.inlines
        new_line = False
        for current, entering in nodes:
            # print(current, current.literal)
            node_type = current.t
            if node_type in ("html_inline", "html_block", "text"):
                context.on_text(current.literal)
            elif node_type == "softbreak":
                if entering:
                    context.on_text("\n")
            elif node_type in inlines:
                if current.is_container():
                    if entering:
                        context.enter_style(f"markdown.{node_type}")
                    else:
                        context.leave_style()
                else:
                    context.enter_style(f"markdown.{node_type}")
                    if current.literal:
                        context.on_text(current.literal)
                    context.leave_style()
                if current.destination and not entering:
                    context.on_text(" (")
                    context.enter_style("markdown.link_url")
                    context.on_text(current.destination)
                    context.leave_style()
                    context.on_text(") ")
            else:
                element_class = self.elements.get(node_type) or UnknownElement
                if current.is_container():
                    if entering:
                        element = element_class.create(self, current)
                        context.stack.push(element)
                        element.on_enter(context)
                    else:
                        element = context.stack.pop()
                        if context.stack:
                            if context.stack.top.on_child_close(context, element):
                                if new_line:
                                    yield Segment("\n")
                                yield from console.render(element, context.options)
                                element.on_leave(context)
                            else:
                                element.on_leave(context)
                        else:
                            element.on_leave(context)
                            yield from console.render(element, context.options)
                        new_line = element.new_line
                else:
                    element = element_class.create(self, current)

                    context.stack.push(element)
                    element.on_enter(context)
                    if current.literal:
                        element.on_text(context, current.literal.rstrip())
                    context.stack.pop()
                    if new_line:
                        yield Segment("\n")
                    yield from console.render(element, context.options)
                    element.on_leave(context)
                    new_line = element.new_line


if __name__ == "__main__":  # pragma: no cover

    markup = """
An h1 header
============
Paragraphs are separated by a blank line.
2nd paragraph. *Italic*, **bold**, and `monospace`. Itemized lists look like:
* this one
* that one
* the other one
Note that --- not considering the asterisk --- the actual text content starts at 4-columns in.
> Block quotes are
> written like so.
>
> They can span multiple paragraphs,
> if you like.
Use 3 dashes for an em-dash. Use 2 dashes for ranges (ex., "it's all in chapters 12--14"). Three dots ... will be converted to an ellipsis. Unicode is supported. ☺
An h2 header
------------
```python
    @classmethod
    def adjust_line_length(
        cls, line: List[Segment], length: int, style: Style = None
    ) -> List[Segment]:        
        line_length = sum(len(text) for text, _style in line)
        if line_length < length:
            return line[:] + [Segment(" " * (length - line_length), style)]
        elif line_length > length:
            line_length = 0
            new_line: List[Segment] = []
            append = new_line.append
            for segment in line:
                segment_length = len(segment.text)
                if line_length + segment_length < length:
                    append(segment)
                    line_length += segment_length
                else:
                    text, style = segment
                    append(Segment(text[: length - line_length], style))
                    break
            return new_line
        return line
```
    """

    import commonmark

    print(commonmark.commonmark(markup))

    from .console import Console

    console = Console(record=True)
    # print(console.size)

    # markup = "<foo>"
    md = Markdown(markup)

    console.print(md)
    print(console)
    # print(console.render_spans())

    # from .color import Color
    # from .style import Style

    # print(Color.downgrade.cache_info())
    # print(Color.parse.cache_info())
    # print(Color.get_ansi_codes.cache_info())

    # Style.parse("on red")

    # print(Style.parse.cache_info())



from typing import Optional

from .console import Console, ConsoleOptions, ConsoleRenderable, RenderResult


class Constrain:
    def __init__(
        self, renderable: ConsoleRenderable, width: Optional[int] = 80
    ) -> None:
        """Constrain the width of a renderable to a given number of characters.
        
        Args:
            renderable (ConsoleRenderable): A renderable object.
            width (int, optional): The maximum width (in characters) to render. Defaults to 80.
        """
        self.renderable = renderable
        self.width = width

    def __console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        if self.width is None:
            yield self.renderable
        else:
            child_options = options.update(width=min(self.width, options.max_width))
            yield from console.render(self.renderable, child_options)



from typing import Match

import re

from ._emoji_codes import EMOJI


def _emoji_replace(text: str, _emoji_sub=re.compile("(:(.*?):)").sub) -> str:
    """Replace emoji code in text."""
    get_emoji = EMOJI.get

    def do_replace(match: Match[str]) -> str:
        """Called by re.sub to do the replacement."""
        emoji_code, emoji_name = match.groups()
        return get_emoji(emoji_name.lower(), emoji_code)

    return _emoji_sub(do_replace, text)
